import time
import os
import requests
import re
from bs4 import BeautifulSoup as BS
from pymongo import MongoClient
from .mongo_engine import *
from .file_operations import *
from datetime import datetime as dt
from datetime import timedelta as td
from .helpers import *
from .decorators.decorators import exponential_backoff, db_lookup

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager

USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.77 Safari/537.36'


class Scraper:

    def __init__(self,
                 db='scraped_data',
                 collection=None,
                 db_host=os.environ.get('MONGODB_URI'),
                 update_interval=None,
                 use_selenium=False,
                 connect_to_db=True):
        '''
        Scraper classes will be tied to the scraped_data db where they will [hopefully] check for data before trying to scrape
        '''
        self.use_selenium = use_selenium
        if connect_to_db:
            self.db = MongoEngine(db, collection, db_host)
        self.update_interval = update_interval or 720  # minutes
        self.UPDATE_INTERVAL = self.update_interval * 60  # seconds
        self.file_operations = FileOperations()
        # self.user_agent = self.set_user_agent()

    @exponential_backoff()
    def make_request(self,
                     url,
                     headers={'Accept-Encoding': 'identity',
                              'User-Agent': USER_AGENT},
                     params={},
                     data={},
                     type='get',
                     timeout=1,
                     verify=True):

        if re.match('https?://', url) == None:
            url = f"https://{url}"

        return requests.get(
            url, headers=headers, params=params, timeout=timeout,
            verify=verify)

    @db_lookup('scraped_data', 'full_html', 'html')
    def get_html_source(self, selection_filter={'url': 'example_url'}):
        '''
        In order for db_lookup decorator to work we need to specify the selection_filter
        used in order to potentially find a cached value for our html

        Simply pass in the parameter like so: \
            {'url': '<the url you want the html from>'}
        '''
        return self.get_html(selection_filter['url'])

    def get_html(self, url, verify=True):
        '''Preference for using selenium because selenium allows the javascript to be run before returning'''

        if self.use_selenium:
            options = Options()
            options.headless = True
            options.add_argument("window-size=1280,800")
            options.add_argument(f"user-agent={USER_AGENT}")
            options.add_argument(
                '--disable-blink-features=AutomationControlled')
            driver = webdriver.Chrome(
                options=options)
            driver.get(url)
            html = driver.page_source
            return html

        else:
            response = self.make_request(url, verify=verify)
            return response.text

    def add_db_entry(self, obj, collection=None, last_updated=True):
        '''
        add obj to collection; if last_updated is true, adds field {'last_updated': dt.now()} to obj
        '''
        if last_updated:
            obj.update({'last_updated': dt.utcnow()})

        self.db[collection].insert_one(obj)

    def get_page_html(self, page_url):
        db_obj = next(self.db["full_html"].find({'page_url': page_url}))
        return db_obj['html'] if(db_obj['created_at'] - dt.now()).seconds < (
            self.update_interval * 60) else self.set_page_html(page_url)

    def set_page_html(self, page_url):
        response = self.make_request(page_url)
        html = response.content.decode('utf-8')
        obj = {'page_url': page_url,
               'html': html}
        self.add_db_entry('full_html', obj)
        return html

    # deprecated cause ugly db_lookup decorator will solve issues of finding specific html, sections wont be stored, only the full html
    # methods for gathering the sections will be defined from the html source which will be generally either updated daily
    # def get_section_html(self, html, section):
    #     db_obj = next(self.db["sections"].find({'section': section}))
    #     return db_obj['html'] if (db_obj['created_at'] - dt.now()).seconds < (self.update_interval * 60) else self.set_section_html(html, section)

    # def set_section_html(self, html, section):
    #     section_regex = f"<a href=\"#{section}\">.*?</a>.*?<ul>(.*?)</ul>"
    #     pattern = re.compile(section_regex, re.S)
    #     section_html = re.findall(pattern, html)[0]
    #     obj = {'section': section,
    #                  'html': section_html}
    #     self.add_db_entry('sections', obj)
    #     return section_html

    def set_user_agent(self, value=None):
        if value:
            self.user_agent = value
        else:
            self.user_agent = get_random_user_agent()

    def rotate_user_agent(self):
        self.user_agent = self.set_user_agent()

    def log_error(self, source, pertinent_info, error):
        logging.basicConfig(
            filename=f'logs/{source}.log', level=logging.INFO,
            format='%(asctime)s %(message)s')
        logging.error(f'Error: {error} caused by line: {pertinent_info}')


class AlphaVantageScraper(Scraper):

    def __init__(self, update_interval=None):
        super().__init__(update_interval=update_interval)

        self.documentation_url = 'https://www.alphavantage.co/documentation/'
        self.html = None

    def get_section_html(self, html, section):
        section_regex = f"<a href=\"#{section}\">.*?</a>.*?<ul>(.*?)</ul>"
        pattern = re.compile(section_regex, re.S)
        section_html = re.findall(pattern, html)[0]
        return section_html

    # @_db_lookup
    def get_tech_indicator_params(self, tech_ind):
        '''
        Returns required and optional fields for passing in a search in form:
            {'required': {<field>: <field_description>, ...},
             'optional': {<field>: <field_description>, ...}
             }
        '''
        tech_ind = tech_ind.lower()

        self.html = self.get_html_source(
            selection_filter={'url': self.documentation_url})

        # response = self.make_request(self.documentation_url)
        # html = response.content.decode('utf-8')
        section_regex = f"<h4 id=\"{tech_ind.replace('_', '')}\">(.*?)Required: <code>apikey"
        pattern = re.compile(section_regex, re.S)
        section = re.findall(pattern, self.html)[0]

        description = re.findall(
            re.compile("<br>.*?<p>(.*?)</p>.*?<br>", re.S),
            section)[0]
        required = re.findall(
            re.compile("Required: ?<code>(.*?)</code>", re.S),
            section)
        optional = re.findall(
            re.compile("Optional: ?<code>(.*?)</code>", re.S),
            section)

        # required descs are the descriptions tied to the required parameters
        required_descs = [
            re.findall(
                re.compile(
                    f"<code>{var}</code></b></p>.*?<p>(.*?)</p>", re.S),
                section)[0].replace('<code>', '').replace('</code>', '')
            for var in required]
        optional_descs = [
            re.findall(
                re.compile(f"<code>{var}</code></p>.*?<p>(.*?)</p>", re.S),
                section)[0].replace('<code>', '').replace('</code>', '')
            for var in optional]

        requirements = zip(required, required_descs)
        options = zip(optional, optional_descs)

        return {
            'description': description,
            'required': dict(requirements),
            'optional': dict(options)
        }

    def get_tech_indicator_params_db(self, tech_ind):
        try:
            return self.db['tech_ind'].find({'name': tech_ind}).next()
        except:
            return None

    def get_high_usage(self, type_="tech_ind", section=''):
        '''
        Pass in a type ("technical-indicators", "time-series-data", "fx" (forex), "digital-currency" (Crypto), "sector-information") \
        to get which indicators are currently being highly used
        '''
        self.get_option_names(type_)[0]

    def get_non_high_usage(self, type="technical-indicators"):
        '''
        Opposite of self.get_high_usage
        '''
        return self.get_option_names(type_)[1]

    def update_db_html(self, page_url):
        pass

    def get_option_names(self, type_="technical-indicators"):
        '''
        type_ options are ("technical-indicators", "time-series-data", "fx" (forex), "digital-currency" (Crypto), "sector-information")

        returns 2 lists: high usage options and low(er) usage options
        '''
        html = self.get_html_source(
            selection_filter={'url': self.documentation_url})

        section = self.get_section_html(html, type_)

        # print(html)
        # print(section)
        options = re.findall(
            re.compile("<li><a href=.*?>([A-Za-z0-9_]+).+?</li>", re.S),
            section)
        high_usage_options = []
        for option in options:
            if len(
                re.findall(
                    re.compile(
                        f"{option} <span.*?(High Usage).*?</li>", re.S),
                    section)) > 0:
                high_usage_options.append(option)

        non_high_usage_options = [
            option for option in options if option not in high_usage_options]

        return high_usage_options, non_high_usage_options

    def get_tech_indicators_json(self):

        output = {}

        high_usage, low_usage = self.get_option_names()

        for tech_ind in high_usage:
            output[tech_ind] = self.get_tech_indicator_params(tech_ind)
            output[tech_ind]['usage'] = 'high'

        for tech_ind in low_usage:
            output[tech_ind] = self.get_tech_indicator_params(tech_ind)
            output[tech_ind]['usage'] = 'low'

        return output

    def set_tech_indicators_json_file(self):

        data = self.get_tech_indicators_json()

        self.file_operations.write_json(data, 'technical-indicators.json')




import configparser
import json
import os
import sys

import slg_utilities.access

CONFIG_FILE = "config.ini"


def get(name):
    """
    Get value of `name` from config prioritized by: runtime, local config, and
    global config. Runtime values are environment variables, local values are
    stored in $HOME/.slgotting/config.json, and global values are stored in
    config.ini in the module installation.
    """
    value = get_runtime_config_value(name)
    if not value:
        value = get_local_config_value(name)
    if not value:
        value = get_global_config_value(name)
    if not value:
        raise ValueError("config value for {} not found".format(name))
    return value

def has_name(name):
    try:
        get(name)
        return True
    except ValueError:
        return False

# Runtime config.
def get_runtime_config_value(name):
    return os.environ.get(name)

def set_runtime_config_value(name, value):
    os.environ[name] = str(value)

# Local config.
def get_local_config_value(name):
    config = read_local_config()
    return config.get(name)

def set_local_config_value(name, value):
    config = read_local_config()
    config[name] = value
    write_local_config(config)

def get_local_config_fpath():
    return os.path.join(os.environ["HOME"], "." + "slgotting", "config.json")

def read_local_config():
    """
    Reads local json config file.
    """
    try:
        config = {}
        config_fpath = get_local_config_fpath()
        if os.path.isfile(config_fpath):
            with open(config_fpath, "r") as fh:
                config = json.load(fh)
        return config
    except Exception as e:
        print("Error reading config. {}".format(str(e)), file=sys.stderr)

def write_local_config(config):
    """
    Writes local json config file.
    """
    try:
        config_fpath = get_local_config_fpath()
        if not os.path.isdir(os.path.dirname(config_fpath)):
            os.makedirs(os.path.dirname(config_fpath))

        with open(config_fpath, "w") as fh:
            json.dump(config, fh, indent=4)
        print("Wrote {}".format(config_fpath))
    except Exception as e:
        print("Error writing config. {}".format(str(e)), file=sys.stderr)

# Global config.
def get_global_config_value(name):
    return read_global_config().get(name)

def get_global_config_fpath():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), CONFIG_FILE)

def read_global_config():
    """
    Returns staging or production global config according to the return value
    of slg_utilities.access.is_production().
    """
    config_fpath = get_global_config_fpath()
    config = configparser.ConfigParser()
    config.read(config_fpath)
    return config["PRODUCTION"] if slg_utilities.access.is_production() \
        else config["STAGING"]



import os
import pymongo
from pymongo import MongoClient
import logging
from datetime import datetime as dt


class MongoEngine:
    '''
    This api will be built around the idea of using a selection_filter to select items in your database, a common task; \
      and using simple understandable function names to operate on the documents returned by the selection_filter

    The purpose is the convenience gained in being able to just say <MongoEngine instance>.remove_key(selection_filter, key) without \
      having to know the weird intricate syntax for simple, common operations

    Names therefore will try to make as much sense as possible (obviously); Coming from my own perspective (which may at any time be ignorant/incomplete)

    Methods will also generally all have a <how_many> parameter, in order to specify if you wish to perform on only one or all \
      of the documents found by the selection_filter

      Perhaps a usage for adding support for n number of documents will present itself but for now \
        one or all seems like the most used cases

    Philosophy will generally be to make easy those functions that are commonplace, and then to not \
      weigh down the software with excessive wrapping of rare usage functionality.

      Use the simple examples and their syntax as documentation for more complex, specialized requirements
    '''

    def __init__(
        self, db='events', collection='events',
            host="mongodb://localhost:27017"):

        logging.basicConfig(
            filename='/var/log/usr/mongo.log', level=logging.INFO,
            format='%(asctime)s %(message)s')

        try:
            self.client = MongoClient(host)
        except Exception as e:
            logging.info(
                f"Could not connect to MongoDB because of error:\n      {e}")

        self.db = self.client[db]
        if collection:
            self.collection = self.db[collection]
        else:
            self.collection = None

    @classmethod
    def establish_connection(cls, host):
        return cls(host=host)

    @property
    def collection_size(self):
        return self.collection.count_documents({})

    def set_collection(self, new_coll):
        self.collection = self.db[new_coll]

    def set_db(self, new_db, new_coll=None, ignore_coll=False):
        self.db = self.client[new_db]
        if new_coll == None and ignore_coll == False:
            self.collection = self.db[input(
                'Please enter a collection name you wish to access in this db: ')]
        elif ignore_coll == True:
            pass
        else:
            self.collection = self.db[new_coll]
        print(self.__repr__())

    def retrieve_all_from_coll(self):
        return list(self.collection.find())

    def filter_date_range(start_date, end_date):
        pass

    def filter_by_mongo_args(self, *args):
        self.collection.find(args)

    def add_entry(self, json, last_updated=True):
        if last_updated:
            json.update({'last_updated': dt.utcnow()})
        return self.collection.insert_one(json)

    def update_specific_value(self, key, value, new_value, how_many='all'):
        '''
        DISCLAIMER: This function does NOT use selection_filter general syntax because we are looking at a specific key \
          and its more explicit what we're doing this way

        By default, finds all with {<key>: <value>} in the document and updates the key with the <new_value>

        how many supports one or all, where one would indicate the first one found matching the key value pair
        '''
        if how_many == 'all':
            self.collection.update_many(
                {key: value},
                {'$set': {key: new_value}})
        elif how_many == 'one':
            self.collection.find_one_and_update(
                {key: value}, {'$set': {key: new_value}})

    def update_value(self, selection_filter, key, value, how_many='all'):
        if how_many == 'all':
            self.collection.update_many(
                selection_filter, {'$set': {key, value}})
        elif how_many == 'one':
            self.collection.find_one_and_update(
                selection_filter, {'$set': {key: value}})

    def remove_key(self, selection_filter, key, how_many='all'):
        '''Removes key from document(s) matching selection_filter'''
        if how_many == 'all':
            self.collection.update_many(
                selection_filter, {'$unset': {key: ''}})
        elif how_many == 'one':
            self.collection.find_one_and_update(
                selection_filter, {'$unset': {key: ''}})

    def remove_all(self, json):
        self.collection.delete_many(json)

    def delete_one(self, dict_):
        self.collection.delete_one(dict_)

    def sort_by(self, *args):
        return list(self.collection.find({}).sort(*args))

    def collection_length(self):
        return self.collection.count()

    def first_empty_id(self, i=0):
        while i < self.collection_length():
            if self.collection.find({'id': i}) == None:
                return i
            i += 1
        return i

    def get_every_key_in_collection(self):
        keys = []
        for obj in list(self.collection.find()):
            for key in obj.keys():
                if key not in keys:
                    keys.append(key)

        # keys = [key for key not in obj.keys() for obj in list(self.collection.find())]

        return keys

    def delete_collection(self, collection):
        verify = input(
            'Input collection name if you are sure you wish to delete this collection: ')
        if verify == collection:
            self.db[collection].drop()

    def __repr__(self):
        return f"Database: {self.db.name}; Collection: {self.collection.name}"

    def __str__(self):
        return f"Database: {self.db.name}; Collection: {self.collection.name}"





from .helpers import *

import gridfs
import base64
import io


class ImageHandler:

    '''
    Standard format of images will be a bytestring, from there can format into useable
    stored objects, or decode them back into their base64 image

    Parameters
    ----------

    images: list
            list of image byte arrays (handling for other formats not supported, yet)

    image_filenames: list : optional
            list of image_filenames that can be zipped into 2 tuple format with
            defaults to f"image{idx+1}"


    DISCLAIMER:
            My understanding of image formats is limited and so
            initialization is currently limited to subclass 'ImgProcessorJS'
            ImageHandler.detect_format() will be used in the future to determine how to reduce
            a list of images to a list of image byte arrays
    '''

    def __init__(self, images=[], image_filenames=[]):
        self.images = images
        self.image_filenames = image_filenames

    def detect_format(self):
        '''
        want to detect current to know how to reformat to standard bytestring format
        not implemented because its not immediately evident or important to do so
        '''
        pass

    @classmethod
    def init_from_image_byte_arrays(cls, images, image_filenames=[]):
        '''
        Accepts images as list of byte arrays and optional image_filenames
        '''
        return cls(images, image_filenames)

    def get_images_as_bytestring(self):
        '''
        returns read images which is basically a bytestring, or its fundamental form
        '''
        return [image.read() for image in self.images]

    def get_images_as_gridFs(self):
        '''
        NOT IMPLEMENTED

        returns images as gridFs objects, useful for storing in a mongo database if images are large (> 16 MB)
                else recommended to store as bytestring
        '''
        return [image]

    def get_images_from_bytestrings(self):
        '''
        data is stored in subclasses typically as bytestrings so this converts them back
        to their base64 image
        '''
        return self.get_images_as_base64()

    def get_images_as_base64(self):
        '''
        encodes images as base64
        '''
        return [base64.b64encode(bytestr) for bytestr in self.images]

    def get_email_ready_format(self):
        '''
        returns images formatted as list of two-tuples of form (image_filename or 'image{idx}', image_byte_array)
        '''
        if self.image_filenames and len(self.image_filenames) > 0:
            return list(zip(self.image_filenames, self.images))
        else:
            return list(
                zip(
                    [f"image{idx+1}"
                     for idx in range(len(self.images))],
                    self.images))


class ImgProcessorJS(ImageHandler):
    '''
    Pass in a javascript formdata object as <ajaxRequest> and return images formatted as you wish

    Handles multiple images by default

when you get the request from ajax, instantiate like so:

image_handler = ImgProcessorJS(request)

the request expects only images, passing other file types will break this


    Make ajax call like so (for images only):

            function passImage() {
                    let file = document.getElementById('file').files[0]
                    var data = new FormData();
                    data.append("img", file);

                    let url = '/example'

                    $.ajax({
                            url: url,
                            data: data,
                            type: 'POST',
                            processData: false,
                            contentType: false,
                            success: function(results) {
                                    prnt(results, 'passed');
                            },
                            error: function(result) {
                                    prnt(result, 'failed')
                            }
                    });


    If passing more data than images, it will look something like this (have to serialize
    other data before adding it to the formdata obj):

            var data = new FormData();
                    var url = "/example"
                    var form_data = $('#user-form').serializeArray();
                    $.each(form_data, function (key, input) {
                                    data.append(input.name, input.value);
                    });
                    var file_data = $('#file')[0].files;
                    data.append("file", file_data[0]);

                    $.ajax({
                            url: url,
                            data: data,
                            type: 'POST',
                            processData: false,
                            contentType: false,
                            success: function(results) {
                                    $('.form-error').each(function() {
                                            $(this).text('');
                                    });
                                    if (results.errors) {
                                            for (const [key, value] of Object.entries(results.errors)) {
                                                    $(`#${key}-error`).text(value)
                                            }
                                    } else {
                                            location.reload(true);
                                    }
                            }
                    });

    '''

    def __init__(self, ajaxRequest):

        self.images = [ajaxRequest.files[key] for key in ajaxRequest.files]
        prnt(type(self.images[0]), 'python image type from ajax files')
        prnt(dir(self.images[0]), 'directory')
        prnt(self.images[0].stream, 'stream')
        self.image_filenames = [image.filename for image in self.images]
        prnt(self.image_filenames, 'image names')

        self.images = self.get_images_as_bytestring()
